/*
 * 
 *  This file is part of Sourceprojects.org Lycia Copyright 2010 noctarius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * 
 */
package org.sourceprojects.lycia.internal.generator;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.xml.XMLConstants;

import org.sourceprojects.lycia.DataConverter;
import org.sourceprojects.lycia.DefinitionBuilder;
import org.sourceprojects.lycia.LyciaParser;
import org.sourceprojects.lycia.ParserContext;
import org.sourceprojects.lycia.QualifiedName;
import org.sourceprojects.lycia.annotations.Attribute;
import org.sourceprojects.lycia.annotations.Base64;
import org.sourceprojects.lycia.annotations.Child;
import org.sourceprojects.lycia.annotations.ContextObject;
import org.sourceprojects.lycia.annotations.Parent;
import org.sourceprojects.lycia.annotations.Tag;
import org.sourceprojects.lycia.annotations.TextBody;
import org.sourceprojects.lycia.annotations.XPath;
import org.sourceprojects.lycia.exceptions.LyciaConfigurationException;
import org.sourceprojects.lycia.internal.AnnotationCombination;
import org.sourceprojects.lycia.internal.dataconverter.BaseDataConverter;
import org.sourceprojects.lycia.internal.dataconverter.ByteDataConverter;
import org.sourceprojects.lycia.internal.logging.LyciaLoggerWrapper;
import org.sourceprojects.lycia.internal.platform.AttributeNameFindingStrategy;
import org.sourceprojects.lycia.internal.platform.ParameterMatcher;
import org.sourceprojects.lycia.internal.platform.PlatformSupportUtils;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class MethodParameterUtilities {

	private static final LyciaLoggerWrapper LOGGER = LyciaLoggerWrapper
			.getLogger(MethodParameterUtilities.class);

	@SuppressWarnings("unchecked")
	private static final AnnotationCombination[] LEGAL_ANNOTATION_COMBINATIONS = {
			new AnnotationCombination(ContextObject.class),
			new AnnotationCombination(Base64.class),

			new AnnotationCombination(Attribute.class),
			new AnnotationCombination(Attribute.class, Base64.class),

			new AnnotationCombination(TextBody.class),
			new AnnotationCombination(TextBody.class, Base64.class),

			new AnnotationCombination(Parent.class),
			new AnnotationCombination(Parent.class, TextBody.class),
			new AnnotationCombination(Parent.class, TextBody.class,
					Base64.class),
			new AnnotationCombination(Parent.class, Attribute.class),
			new AnnotationCombination(Parent.class, Attribute.class,
					Base64.class),

			new AnnotationCombination(Child.class),
			new AnnotationCombination(Child.class, TextBody.class),
			new AnnotationCombination(Child.class, TextBody.class, Base64.class),
			new AnnotationCombination(Child.class, Attribute.class),
			new AnnotationCombination(Child.class, Attribute.class,
					Base64.class), };

	private MethodParameterUtilities() {
	}

	public static <B> Object[] generateParameter(final Method method,
			final Element element, final String localName,
			final ParserContext<B> parserContext,
			final DefinitionBuilder definitionBuilder,
			final LyciaParser<B> lyciaParser) throws Exception {

		final Class<?>[] parameters = method.getParameterTypes();
		final Annotation[][] annotations = method.getParameterAnnotations();

		final Object[] result = new Object[parameters.length];

		for (int i = 0; i < parameters.length; i++) {
			final Class<?> parameter = parameters[i];

			final Annotation[] reodered = precheckAndReorderAnnotations(method,
					i, annotations[i], lyciaParser);

			Object value = null;
			if (reodered.length > 0) {
				for (final Annotation annotation : reodered) {
					final ParameterMatcher matcher = getParameterMatcher(
							parameter, annotation, lyciaParser);

					if (matcher != null) {
						value = getParameterObjectByType(parameter, annotation,
								matcher, value, element, localName,
								parserContext, definitionBuilder, lyciaParser);
					}
				}

				result[i] = value;
			} else {
				final ParameterMatcher matcher = getParameterMatcher(parameter,
						null, lyciaParser);

				result[i] = getParameterObjectByType(parameter, null, matcher,
						null, element, localName, parserContext,
						definitionBuilder, lyciaParser);
			}
		}

		return result;
	}

	public static <B> Annotation[] precheckAndReorderAnnotations(
			final Method method, final int index,
			final Annotation[] annotations, final LyciaParser<B> lyciaParser) {

		if (annotations.length == 0) {
			return new Annotation[0];
		}

		final Annotation[] result = new Annotation[annotations.length];

		int pos = 0;
		final List<Annotation> lyciaAnnotations = new ArrayList<Annotation>();
		for (int i = 0; i < annotations.length; i++) {
			Annotation annotation = annotations[i];

			if (!isLyciaParameterAnnotation(annotation)) {
				result[pos++] = annotation;
			} else {
				if (Attribute.class.equals(annotation.annotationType())) {
					lyciaAnnotations.add(buildAttributeByNamingStrategy(method,
							index, (Attribute) annotation, lyciaParser));
				} else {
					lyciaAnnotations.add(annotation);
				}
			}
		}

		final Annotation[] combination = reorderAnnotations(lyciaAnnotations);
		for (final Annotation annotation : combination) {
			result[pos++] = annotation;
		}

		return result;
	}

	public static <B> ParameterMatcher getParameterMatcher(
			final Class<?> parameter, final Annotation annotation,
			final LyciaParser<B> lyciaParser) {

		for (final ParameterMatcher matcher : lyciaParser.getConfigurator()
				.getPlatformFactory().newParameterMatcher()) {

			if (matcher.getParamaeterClass().isAssignableFrom(parameter)) {
				if (matcher.getAnnotationClass() == null) {
					return matcher;
				}

				if (matcher.getAnnotationClass() != null && annotation != null) {
					if (matcher.getAnnotationClass().equals(
							annotation.annotationType())) {
						return matcher;
					}
				}
			}
		}

		return null;
	}

	public static <B> Object getParameterObjectByType(final Class<?> parameter,
			final Annotation annotation, final ParameterMatcher matcher,
			final Object value, final Element element, final String localName,
			final ParserContext<B> parserContext,
			final DefinitionBuilder definitionBuilder,
			final LyciaParser<B> lyciaParser) throws Exception {

		if (matcher.getAnnotationClass() == null) {
			if (Element.class.isAssignableFrom(parameter)) {
				return element;
			}

			if (String.class.isAssignableFrom(parameter)) {
				return localName;
			}

			if (QualifiedName.class.isAssignableFrom(parameter)) {
				return new QualifiedName(element.getNamespaceURI(), element
						.getLocalName());
			}

			if ("javax.xml.namespace.QName"
					.equals(parameter.getCanonicalName())) {

				final Class<?> qNameClass = Class
						.forName("javax.xml.namespace.QName");
				final Constructor<?> constructor = qNameClass.getConstructor(
						String.class, String.class);

				return constructor.newInstance(element.getNamespaceURI(),
						element.getLocalName());
			}

			if (ParserContext.class.isAssignableFrom(parameter)) {
				return parserContext;
			}

			if (DefinitionBuilder.class.isAssignableFrom(parameter)) {
				return definitionBuilder;
			}
		}

		final Class<? extends Annotation> clazz = matcher.getAnnotationClass();
		if (clazz.equals(Child.class)) {
			LOGGER.debug("Found @Child annotation to inject");
			return getChildElement(element, annotation, parameter);

		} else if (clazz.equals(Parent.class)) {
			LOGGER.debug("Found @Parent annotation to inject");
			return getParentElement(element, parameter);

		} else if (clazz.equals(ContextObject.class)) {
			LOGGER.debug("Found @ContextObject annotation to inject");
			return parserContext.getContextObject();

		} else if (clazz.equals(Base64.class)) {
			LOGGER.debug("Found @Base64 annotation to inject");
			final DataConverter converter = ByteDataConverter.BASE_BYTE_DATA_CONVERTER;
			final Object encoded = (value != null ? value : getTextBody(
					element, lyciaParser));
			return converter.convert(encoded, null);

		} else if (clazz.equals(TextBody.class)) {
			LOGGER.debug("Found @TextBody annotation to inject");
			return getTextBody(element, lyciaParser);

		} else if (clazz.equals(Attribute.class)) {
			LOGGER.debug("Found @Attribute annotation to inject");
			return getConvertedAttributeValue(parameter, annotation, element);
		}

		return null;
	}

	private static <B> String getTextBody(final Element element,
			final LyciaParser<B> lyciaParser) {

		return lyciaParser.getConfigurator().getPlatformFactory()
				.newElementContentExtractor().getElementContent(element);
	}

	private static boolean isExtendedAttribute(
			final Class<? extends Annotation> clazz) {
		if (clazz.getAnnotation(Attribute.class) != null) {
			return true;
		}

		return false;
	}

	private static Object getParentElement(final Element element,
			final Class<?> type) {

		if (Element.class.isAssignableFrom(type)) {
			return element.getParentNode();
		}

		if (String.class.isAssignableFrom(type)) {
			if (element.getParentNode() != null) {
				return element.getParentNode().getLocalName();
			}

			return null;
		}

		if (QualifiedName.class.isAssignableFrom(type)) {
			if (element.getParentNode() != null) {
				final Node parent = element.getParentNode();
				return new QualifiedName(parent.getNamespaceURI(), parent
						.getLocalName());
			}

			return null;
		}

		if ("javax.xml.namespace.QName".equals(type.getCanonicalName())) {
			if (!PlatformSupportUtils.isDalvikDependingPlatform()) {
				try {
					final Class<?> qNameClass = Class
							.forName("javax.xml.namespace.QName");
					final Constructor<?> constructor = qNameClass
							.getConstructor(String.class, String.class);

					return constructor.newInstance(element.getNamespaceURI(),
							element.getLocalName());
				} catch (final ClassNotFoundException e) {
					throw new LyciaConfigurationException(e);
				} catch (final SecurityException e) {
					throw new LyciaConfigurationException(e);
				} catch (final NoSuchMethodException e) {
					throw new LyciaConfigurationException(e);
				} catch (final IllegalArgumentException e) {
					throw new LyciaConfigurationException(e);
				} catch (final InstantiationException e) {
					throw new LyciaConfigurationException(e);
				} catch (final IllegalAccessException e) {
					throw new LyciaConfigurationException(e);
				} catch (final InvocationTargetException e) {
					throw new LyciaConfigurationException(e);
				}
			}
		}

		throw new LyciaConfigurationException(
				"@Parent annotation is used at illegal parameter-type. "
						+ "Only legal on String, QualifiedName, QName and Element");
	}

	private static Object getChildElement(final Element element,
			final Annotation annotation, Class<?> type) {
		final Element result = (Element) getChildElementByExpression(element,
				annotation);

		if (result == null) {
			return null;
		}

		if (Element.class.isAssignableFrom(type)) {
			return result;
		}

		if (String.class.isAssignableFrom(type)) {
			if (element.getParentNode() != null) {
				return result.getLocalName();
			}

			return null;
		}

		if (QualifiedName.class.isAssignableFrom(type)) {
			return new QualifiedName(result.getNamespaceURI(), result
					.getLocalName());
		}

		if ("javax.xml.namespace.QName".equals(type.getCanonicalName())) {
			if (!PlatformSupportUtils.isDalvikDependingPlatform()) {
				try {
					final Class<?> qNameClass = Class
							.forName("javax.xml.namespace.QName");
					final Constructor<?> constructor = qNameClass
							.getConstructor(String.class, String.class);

					return constructor.newInstance(result.getNamespaceURI(),
							result.getLocalName());

				} catch (final ClassNotFoundException e) {
					throw new LyciaConfigurationException(e);
				} catch (final SecurityException e) {
					throw new LyciaConfigurationException(e);
				} catch (final NoSuchMethodException e) {
					throw new LyciaConfigurationException(e);
				} catch (final IllegalArgumentException e) {
					throw new LyciaConfigurationException(e);
				} catch (final InstantiationException e) {
					throw new LyciaConfigurationException(e);
				} catch (final IllegalAccessException e) {
					throw new LyciaConfigurationException(e);
				} catch (final InvocationTargetException e) {
					throw new LyciaConfigurationException(e);
				}
			}
		}

		throw new LyciaConfigurationException(
				"@Child annotation is used at illegal parameter-type. "
						+ "Only legal on String, QualifiedName, QName and Element");
	}

	private static Object getChildElementByExpression(final Element element,
			final Annotation annotation) {

		if (!element.hasChildNodes()) {
			return null;
		}

		final Tag tag = ((Child) annotation).value();

		if (tag.xpath() != null && !tag.xpath().expression().equals(XPath.NULL)) {
			throw new LyciaConfigurationException(
					"XPath expressions aren't supported on @Child constructs");
		}

		if (tag.regex()) {
			throw new LyciaConfigurationException(
					"Regex expressions aren't supported on @Child constructs");
		}

		final NodeList nodes = element.getChildNodes();
		for (int i = 0; i < nodes.getLength(); i++) {
			final Node node = nodes.item(i);

			if (node.getNodeType() != Node.ELEMENT_NODE) {
				continue;
			}

			final Element el = (Element) node;

			if (el.getLocalName().equals(tag.value())
					&& (el.getNamespaceURI() == null
							&& tag.namespace().equals(XMLConstants.NULL_NS_URI) || el
							.getNamespaceURI() != null
							&& el.getNamespaceURI().equals(tag.namespace()))) {
				return el;
			}
		}

		return null;
	}

	private static <B> Object getConvertedAttributeValue(
			final Class<?> parameter, final Annotation annotation,
			final Element element) throws Exception {

		final Attribute attribute = (Attribute) annotation;

		DataConverter converter;
		if (attribute.converter().equals(DataConverter.class)) {
			converter = BaseDataConverter.BASE_DATA_CONVERTER;
		} else {
			final Class<? extends DataConverter> converterClass = attribute
					.converter();
			converter = converterClass.newInstance();
		}

		// Workaround for IBM JVMs
		final Attr attr = matchAttribute(element, attribute.namespaceURI(),
				attribute.value());

		if (attr == null) {
			return null;
		}

		if (LOGGER.isDebugEnabled()) {
			LOGGER.debug("Attribute: " + attr);
			LOGGER.debug("Final parametertype: " + parameter);
			LOGGER.debug("Converter instance: " + converter);
		}

		return converter.convert(attr.getValue(), parameter);
	}

	private static Attr matchAttribute(final Element element,
			String namespaceURI, final String localName) {

		if (namespaceURI.equals(XMLConstants.NULL_NS_URI)) {
			namespaceURI = null;
		}

		final NamedNodeMap map = element.getAttributes();
		for (int i = 0; i < map.getLength(); i++) {
			final Node node = map.item(i);
			LOGGER.debug("Node found: " + node.getNamespaceURI() + "=>"
					+ node.getLocalName());

			if (node.getLocalName().equals(localName)) {
				if (node.getNamespaceURI() == null && namespaceURI == null
						|| node.getNamespaceURI().equals(namespaceURI)) {
					return (Attr) node;
				}
			}
		}

		return null;
	}

	private static Annotation[] reorderAnnotations(final List<Annotation> list) {
		Annotation[] annotations = new Annotation[list.size()];
		annotations = list.toArray(annotations);

		for (final AnnotationCombination combination : LEGAL_ANNOTATION_COMBINATIONS) {
			if (isLegalCombination(combination, annotations)) {
				final Annotation[] result = new Annotation[annotations.length];

				final Class<? extends Annotation>[] types = combination
						.getAnnotations();

				for (int i = 0; i < types.length; i++) {
					result[i] = getAnnotationByType(annotations, types[i]);
				}

				return result;
			}
		}

		final StringBuilder sb = new StringBuilder(
				"An illegal annotation combination was found [");
		sb.append(Arrays.toString(annotations)).append(']');
		throw new LyciaConfigurationException(sb.toString());
	}

	private static boolean isLegalCombination(
			final AnnotationCombination combination,
			final Annotation[] annotations) {

		if (annotations.length != combination.getAnnotations().length) {
			return false;
		}

		return recursiveIsLegalCombination(combination.getAnnotations(),
				annotations, 0);
	}

	private static boolean recursiveIsLegalCombination(
			final Class<? extends Annotation>[] types,
			final Annotation[] annotations, int index) {

		final Class<? extends Annotation> type = annotations[index]
				.annotationType();

		if (types[index].equals(type) || types[index].equals(Attribute.class)
				&& isExtendedAttribute(type)) {
			if (++index == annotations.length) {
				return true;
			}

			return recursiveIsLegalCombination(types, annotations, index);
		}

		return false;
	}

	private static boolean isLyciaParameterAnnotation(
			final Annotation annotation) {
		final Class<? extends Annotation> type = annotation.annotationType();

		if (type.equals(Child.class)) {
			return true;
		}

		if (type.equals(Parent.class)) {
			return true;
		}

		if (type.equals(Base64.class)) {
			return true;
		}

		if (type.equals(Attribute.class)) {
			return true;
		}

		if (isExtendedAttribute(type)) {
			return true;
		}

		if (type.equals(TextBody.class)) {
			return true;
		}

		if (type.equals(ContextObject.class)) {
			return true;
		}

		return false;
	}

	private static <B> Attribute buildAttributeByNamingStrategy(
			final Method method, final int index, final Attribute attribute,
			final LyciaParser<B> lyciaParser) {

		final AttributeNameFindingStrategy<B> strategy = lyciaParser
				.getConfigurator().getPlatformFactory()
				.newAttributeNameFindingStrategy();

		final String name = strategy.findAttributeName(method, index);

		return new Attribute() {
			@Override
			public Class<? extends Annotation> annotationType() {
				return Attribute.class;
			}

			@Override
			public String value() {
				return name;
			}

			@Override
			public String namespaceURI() {
				return attribute.namespaceURI();
			}

			@Override
			public Class<? extends DataConverter> converter() {
				return attribute.converter();
			}
		};
	}

	@SuppressWarnings("unchecked")
	private static <A extends Annotation> A getAnnotationByType(
			final Annotation[] annotations, final Class<A> annotationClass) {

		for (final Annotation a : annotations) {
			if (a.annotationType().equals(annotationClass)) {
				return (A) a;
			} else if (isExtendedAttribute(a.annotationType())) {
				LOGGER.debug("Found extended @Attribute annotation '"
						+ a.annotationType().getName() + "'");
				return (A) a.annotationType().getAnnotation(Attribute.class);
			}
		}

		return null;
	}

}
